#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:00 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

from PIL import Image
import io
import os
import csv
import h5py
import numpy as np

from .decorator import Timer

class FuelGenerator(object):

    def __init__(self, config=None, file_dir=None, file_list=None, dtype=None,
                 height=0, width=0, channel=1, encode='jpeg', is_seq=False):
        self.param = {}
        self.data = {}
        self.modifiable_keys = ['file_dir', 'file_list', 'dtype',
                                'height', 'width', 'channel', 'encode']
        self.valid_dtypes = ['uint8', 'uint16', 'uint32', 'int8', 'int16', 'int32', 'int64',
                             'float16', 'float32', 'float64', 'str', 'bool']
        if config:
            self.param['file_dir'] = config.get('file_dir')
            self.param['file_list'] = config.get('file_list')
            self.param['dtype'] = [np.dtype(dt) for dt in config.get('dtype')]
            self.param['height'] = config.get('height', height)
            self.param['width'] = config.get('width', width)
            self.param['channel'] = config.get('channel', channel)
            self.param['encode'] = config.get('encode', encode)
            self.param['is_seq'] = config.get('is_seq', encode)
        else:
            self.param['file_dir'] = file_dir
            self.param['file_list'] = file_list
            self.param['dtype'] = dtype
            self.param['height'] = height
            self.param['width'] = width
            self.param['channel'] = channel
            self.param['encode'] = encode
            self.param['is_seq'] = is_seq

        # check if key arguments are valid
        if self.param['file_list'].split('.')[-1] != 'csv':
            raise Exception('file list should be a csv file.')
        for dt in self.param['dtype']:
            if dt not in self.valid_dtypes:
                raise Exception('%s is not a valid data type.' % dt)


    def _preProcess(self, img_path):
        ch_err = Exception('Images having %d channels are invalid.' % self.param['channel'])
        if self.param['channel'] == 1:
            image = Image.open(img_path).convert('L')
        elif self.param['channel'] == 3:
            image = Image.open(img_path)
        else:
            raise ch_err
        if self.param['width']>0 and self.param['height']>0:
            image = image.resize((self.param['width'], self.param['height']))
        with io.BytesIO() as buffer:
            image.save(buffer, format=self.param['encode'])
            encoded_bytes = buffer.getvalue()
        np_bytes = np.frombuffer(encoded_bytes, dtype=np.uint8)

        return np_bytes

    @Timer
    def _file2Byte(self, dst_path):
        print('+' + (60 * '-') + '+')
        nsample = len(open(os.path.join(self.param['file_dir'], self.param['file_list']), 'r').readlines()) - 1
        with open(os.path.join(self.param['file_dir'], self.param['file_list']), 'r') as filelist:
            content = csv.reader(filelist)
            ending_char = '\r'
            for l, line in enumerate(content):
                # display progress bar
                progress = int(l/nsample*19 + 0.4)
                yellow_bar = progress * ' '
                space_bar = (19-progress) * ' '
                if l == nsample:
                    ending_char = '\n'
                print('| Integrating data %7d / %-7d ⊰⟦\033[43m%s\033[0m%s⟧⊱ |'
                      % (l, nsample, yellow_bar, space_bar), end=ending_char)
                if l == 0: # initialize data dict
                    self.info_keys = line
                    if len(line) != len(self.param['dtype']):
                        raise Exception('number of given dtypes does not match the provided csv file.')
                    for key in line:
                        self.data[key] = []
                    max_frames = 0
                else:
                    for k, key in enumerate(self.info_keys):
                        if k == 0: # dealing with raw data
                            if self.param['is_seq']:
                                csl = line[k].split(',')  # comma separated line
                                max_frames = len(csl) if len(csl)>max_frames else max_frames
                                temp_data = []
                                for f in csl:
                                    temp_data.append(self._preProcess(os.path.join(self.param['file_dir'], f)))
                                self.data[key].append(temp_data)
                            else:
                                self.data[key].append(self._preProcess(os.path.join(self.param['file_dir'], line[k])))
                        else:
                            csl = line[k].split(',') # comma separated line
                            if len(csl) == 1:
                                self.data[key].append(line[k])
                            else:
                                temp_data = []
                                for f in csl:
                                    temp_data.append(f)
                                self.data[key].append(temp_data)
        hdf5 = h5py.File(dst_path, 'w')
        for k, key in enumerate(self.info_keys):
            if k == 0: # dealing with raw data
                dt = h5py.special_dtype(vlen=self.param['dtype'][k])
                hdf5.create_dataset(key, dtype=dt, data=np.array(self.data[key]))
            else:
                hdf5[key] = np.array(self.data[key]).astype(self.param['dtype'][k])
        if self.param['is_seq']:
            hdf5['max_frames'] = max_frames
        hdf5.close()

    def generateFuel(self, dst_path):
        if dst_path.split('.')[-1] != 'hdf5':
            raise Exception('hdf5 file is recommended for storing compressed data.')
        duration = self._file2Byte(dst_path)
        print('+' + (60 * '-') + '+')
        print('| \033[1;35m%-23s\033[0m has been generated within \033[1;35m%7.2fs\033[0m |'
              % (os.path.basename(dst_path), duration))
        print('+' + (60 * '-') + '+')

    def editProperty(self, config=None, **kwargs):
        if config:
            kwargs = config
        for key in kwargs:
            if key not in self.modifiable_keys:
                print('%s is not a modifiable parameter or has not been defined.' % key)
            else:
                self.param[key] = kwargs[key]


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 2:22 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========

'''
# -*- coding:utf-8 -*-

import nebulae.toolkit.fuel_generator
import nebulae.fuel.fuel_depot
import nebulae.fuel.fuel_tank
import nebulae.spacedock.space_craft
import nebulae.spacedock.component
import nebulae.cockpit.engine
import nebulae.aerolog.layout_sheet

name = 'nebulae'
__all__ = ['fuel', 'spacedock', 'cockpit', 'aerolog', 'toolkit']


#!/usr/bin/env python
'''
space_craft
Created by Seria at 23/11/2018 10:31 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

import tensorflow as tf
import numpy as np
from .component import Pod

class SpaceCraft(object):
    def __init__(self, layout_sheet, scope='', reuse=None, verbose=True):
        assert '/' not in scope
        if scope:
            self.scope = scope + '/'
        else:
            self.scope = scope
        if reuse is None:
            self.reuse = tf.AUTO_REUSE
        self.mile = tf.Variable(0, trainable=False)
        self.valid_dtypes = ['uint8', 'uint16', 'uint32', 'int8', 'int16', 'int32', 'int64',
                             'float16', 'float32', 'float64', 'str', 'bool']
        self.operands = {'+': tf.add, '-': tf.subtract, '*': tf.multiply, '@': tf.matmul, '&': tf.concat}
        self.operands_counter = {'+': 0, '-': 0, '*': 0, '@': 0, '&': 0}
        self.layout = {}
        self.layout_sheet = layout_sheet
        self.verbose = verbose

    def fuelLine(self, name, shape, dtype, default=None):
        if dtype not in self.valid_dtypes:
            raise Exception('%s is not a valid data type.' % dtype)
        if not default is None:
            self.layout[name] = tf.placeholder_with_default(tf.constant(default, dtype=dtype),
                                                            shape, 'FL/DEFAULT_' + name)
        else:
            self.layout[name] = tf.placeholder(tf.as_dtype(dtype), shape, 'FL/' + name)

    def _getHull(self, component):
        if isinstance(component, (tf.Tensor, tf.SparseTensor, tf.Variable)):
            tshape = component.get_shape()
            if tshape.dims is None or tshape.ndims == 0:
                return ' 1    '
            shape = tshape[1:].as_list()
        elif isinstance(component, np.ndarray):
            shape = list(component.shape)
        else:
            shape = [1]
        len_wo_bs = len(shape) - 1
        hull = ' '
        for dim in range(len_wo_bs):
            hull += '%-4d x ' % shape[dim]
        hull += '%-5d' % shape[-1]
        return hull

    def _assemble(self, left_comp, len_prefix, right_comp=None, gear='fit', assemblage=None, sub_scope=''):
        left_symbol = left_comp.symbol
        left_msg = left_comp.message
        if sub_scope:
            sub_prefix = sub_scope + '/'
        else:
            sub_prefix = sub_scope
        if gear == 'fit':
            # this is an atomic component
            if not isinstance(left_symbol, str) or left_symbol.startswith('.'):
                node_name = sub_prefix + left_comp.name
                if node_name in self.layout.keys():  # Reuse or Redefine an existent node
                    hull = self._getHull(left_symbol)
                    self.layout_sheet._drawNode(left_symbol.op.name[len_prefix:], node_name, hull)
                    return sub_prefix+left_comp.name
                if assemblage is None:
                    assert not (isinstance(left_symbol, str) and isinstance(left_msg, str))
                    self.layout[node_name] = left_comp.component(name=left_comp.name)
                    # draw layout sheet
                    if isinstance(left_symbol, (tf.Tensor, tf.SparseTensor, tf.Variable)):
                        hull = self._getHull(left_symbol)
                        if left_symbol.name.startswith('FL/'):  # if the input is a fuel line
                            self.layout_sheet._drawNode(left_symbol.op.name[3:], node_name,
                                                        hull, init=True)
                        else:
                            self.layout_sheet._drawNode(left_symbol.op.name[len_prefix:],
                                                        node_name, hull)
                    for msg in left_msg:
                        if isinstance(msg, (tf.Tensor, tf.SparseTensor, tf.Variable)):
                            hull = self._getHull(msg)
                            if msg.name.startswith('FL/'):  # if the input is a fuel line
                                self.layout_sheet._drawNode(msg.op.name[3:], node_name,
                                                            hull, init=True)
                            else:
                                self.layout_sheet._drawNode(msg.op.name[len_prefix:],
                                                            node_name, hull)
                else:
                    if left_symbol[1:] == 'penalty':
                        self.layout[node_name] = left_comp.component(name=left_comp.name)
                        assemblage = None
                    else:
                        self.layout[node_name] = left_comp.component(name=left_comp.name,
                                                                     input=self.layout[assemblage])
                    # draw layout sheet
                    if assemblage:
                        hull = self._getHull(self.layout[assemblage])
                        self.layout_sheet._drawNode(assemblage, node_name, hull)
                    if not isinstance(left_symbol, str):
                        hull = self._getHull(left_symbol)
                        if left_symbol.name.startswith('FL/'):  # if the input is a fuel line
                            self.layout_sheet._drawNode(left_symbol.op.name[3:], node_name,
                                                        hull, init=True)
                        else:
                            hull = self._getHull(self.layout[assemblage])
                            self.layout_sheet._drawNode(left_symbol.op.name[len_prefix:],
                                                        node_name, hull)
                    for msg in left_msg:
                        if isinstance(msg, (tf.Tensor, tf.SparseTensor, tf.Variable)):
                            hull = self._getHull(msg)
                            if msg.name.startswith('FL/'):  # if the input is a fuel line
                                self.layout_sheet._drawNode(msg.op.name[3:], node_name,
                                                            hull, init=True)
                            else:
                                self.layout_sheet._drawNode(msg.op.name[len_prefix:],
                                                            node_name, hull)
                return sub_prefix+left_comp.name
            elif left_symbol == '>':
                for comp in left_comp.component:
                    assemblage = self.assemble(comp, assemblage=assemblage,
                                               sub_scope=sub_scope, is_external=False)
                return assemblage
            elif left_symbol in ['+', '-', '*', '@', '&']:
                operator = None
                self.operands_counter[left_symbol] += 1
                # (TODO) if a pod is extended to more than two comps, assemblage_name needs modification
                assemblage_name = left_comp.name + '_' + str(self.operands_counter[left_symbol])
                assemblage_list = []
                hull_list = []
                init_assemblage = assemblage
                for comp in left_comp.component:
                    assemblage = self.assemble(comp, assemblage=init_assemblage,
                                               sub_scope=sub_scope, is_external=False)
                    assemblage_list.append(assemblage)
                    if self.verbose:
                        hull = self._getHull(self.layout[assemblage])
                        hull_list.append(hull)
                    if operator is None:
                        operator = self.layout[assemblage]
                    else:
                        if left_symbol == '&':
                            operator = self.operands[left_symbol]([operator, self.layout[assemblage]],
                                                                  axis=-1, name=assemblage_name)
                        else:
                            operator = self.operands[left_symbol](operator,
                                                                  self.layout[assemblage],
                                                                  name=assemblage_name)
                self.layout[sub_prefix + assemblage_name] = operator
                self.layout_sheet._combineNodes(assemblage_list, hull_list,
                                                sub_prefix + assemblage_name, left_symbol)
                return sub_prefix+assemblage_name
            else:
                raise TypeError('unsupported operand type for %s: "Pod" and "Pod".' % left_comp.symbol)
                # elif gear == ''

    def assemble(self, left_comp, right_comp=None, gear='fit', assemblage=None, sub_scope='', is_external=True):
        if not isinstance(left_comp, Pod):
            raise TypeError('Only the members from Component can be assembled.')
        if self.scope[:3] == 'FL/':
            raise ValueError('FL is the reserved words for fuel lines in spacecraft. '
                             + 'Please change your global scope name.')
        len_prefix = len(self.scope)
        if not is_external:
            return self._assemble(left_comp, len_prefix, assemblage=assemblage, sub_scope=sub_scope)
        else:
            if sub_scope:
                scope = self.scope+sub_scope+'/'
            else:
                scope = self.scope
            with tf.variable_scope(scope, reuse=self.reuse):
                return self._assemble(left_comp, len_prefix, assemblage=assemblage, sub_scope=sub_scope)


#!/usr/bin/env python
'''
setup
Created by Seria at 04/11/2018 10:50 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nebulae",
    version="0.1.4",
    author="Seria",
    author_email="zzqsummerai@yeah.net",
    description="A novel and simple framework based on prevalent DL framework and other image processing libs."
                + " v0.1.4: Now the gpu setting will be written in config as it is rather than the detected result.",
                # + "v0.1.1: change the way to register stage; "
                # + "fix a bug in DashBoard may draw points in wrong places.",
                # + "v0.1.0: A roughly complete version. New parts, Engine, Time Machine, Dashboard and Navigator are added.",
                # + "v0.0.18: fix a bug would repeatedly append variable scope.",
                # + "v0.0.17: network layout will be saved as image instead of pdf; "
                # + "add components: RESHAPE, SLICE, CLIP;"
                # + "implement Time Machine.",
                # + "v0.0.15: fix a bug would lead to wrong implementation of ** symbol."
                # + "v0.0.14: unable to assemble DUPLICATE component without name but passing existent name is allowed."
                # + "v0.0.13: allow users to assemble DUPLICATE component without passing name; "
                # + "add RESIZE component; "
                # + "shorten automatically generated component names."
                # + "v0.0.12: fix wrong implementation on data augmentation; "
                # + "FuelGenerator will keep original image size if width or height is not given.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/",
    packages=setuptools.find_packages(),
    install_requires=
    ['graphviz',
     'h5py'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)